# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from continual.rpc.logging.v1 import logging_pb2 as continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class LoggingAPIStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.LogHeartbeat = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogHeartbeat',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogExperiments = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogExperiments',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogMetrics = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogMetrics',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogEvaluationModelMetrics = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogEvaluationModelMetrics',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogDatasetStats = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogDatasetStats',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogPredictionStats = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogPredictionStats',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GetArtifact = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/GetArtifact',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.FromString,
                )
        self.BatchGetArtifacts = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/BatchGetArtifacts',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.FromString,
                )
        self.DeleteArtifact = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/DeleteArtifact',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.GenerateArtifactUploadURL = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/GenerateArtifactUploadURL',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.SerializeToString,
                response_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.FromString,
                )
        self.LogDataChecks = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogDataChecks',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogModelVersion = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogModelVersion',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.LogBatchPrediction = channel.unary_unary(
                '/continual.rpc.logging.v1.LoggingAPI/LogBatchPrediction',
                request_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class LoggingAPIServicer(object):
    """Missing associated documentation comment in .proto file."""

    def LogHeartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogExperiments(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogMetrics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogEvaluationModelMetrics(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogDatasetStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogPredictionStats(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BatchGetArtifacts(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GenerateArtifactUploadURL(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogDataChecks(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogModelVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LogBatchPrediction(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_LoggingAPIServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'LogHeartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.LogHeartbeat,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogExperiments': grpc.unary_unary_rpc_method_handler(
                    servicer.LogExperiments,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogMetrics': grpc.unary_unary_rpc_method_handler(
                    servicer.LogMetrics,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogEvaluationModelMetrics': grpc.unary_unary_rpc_method_handler(
                    servicer.LogEvaluationModelMetrics,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogDatasetStats': grpc.unary_unary_rpc_method_handler(
                    servicer.LogDatasetStats,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogPredictionStats': grpc.unary_unary_rpc_method_handler(
                    servicer.LogPredictionStats,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GetArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.GetArtifact,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.SerializeToString,
            ),
            'BatchGetArtifacts': grpc.unary_unary_rpc_method_handler(
                    servicer.BatchGetArtifacts,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.SerializeToString,
            ),
            'DeleteArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteArtifact,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'GenerateArtifactUploadURL': grpc.unary_unary_rpc_method_handler(
                    servicer.GenerateArtifactUploadURL,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.FromString,
                    response_serializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.SerializeToString,
            ),
            'LogDataChecks': grpc.unary_unary_rpc_method_handler(
                    servicer.LogDataChecks,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogModelVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.LogModelVersion,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'LogBatchPrediction': grpc.unary_unary_rpc_method_handler(
                    servicer.LogBatchPrediction,
                    request_deserializer=continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'continual.rpc.logging.v1.LoggingAPI', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class LoggingAPI(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def LogHeartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogHeartbeat',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogHeartbeatRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogExperiments(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogExperiments',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogExperimentsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogMetrics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogMetrics',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogMetricsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogEvaluationModelMetrics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogEvaluationModelMetrics',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogEvaluationModelMetricsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogDatasetStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogDatasetStats',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDatasetStatsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogPredictionStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogPredictionStats',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogPredictionStatsRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/GetArtifact',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GetArtifactRequest.SerializeToString,
            continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.Artifact.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BatchGetArtifacts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/BatchGetArtifacts',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsRequest.SerializeToString,
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.BatchGetArtifactsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/DeleteArtifact',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.DeleteArtifactRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GenerateArtifactUploadURL(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/GenerateArtifactUploadURL',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLRequest.SerializeToString,
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.GenerateArtifactUploadURLResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogDataChecks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogDataChecks',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogDataChecksRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogModelVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogModelVersion',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogModelVersionRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LogBatchPrediction(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/continual.rpc.logging.v1.LoggingAPI/LogBatchPrediction',
            continual_dot_rpc_dot_logging_dot_v1_dot_logging__pb2.LogBatchPredictionRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import typer
import yaml

from continual.python.cli import utils, feature_sets, model_versions
from continual.python.sdk.client import Client
from continual.rpc.management.v1.types import ModelVersionState

from pathlib import Path
from rich.console import Console
from typing import List

app = typer.Typer(help="Manage models.")


def get_linked_entities(columns):
    entities = {}
    for c in columns:
        entity = c.entity
        if entity is not None and len(entity) > 0:
            entities[c.name] = entity
    return "\n".join(["%s : %s" % (k, v) for k, v in entities.items()])


def format_model_data(m, zipped=False, all_projects=False):
    create_time = m.create_time.replace(microsecond=0)
    update_time = m.update_time.replace(microsecond=0)
    num_bpj = len(m.batch_prediction_jobs.list(1000))
    num_d = len(m.promotions.list(1000))
    num_mv = len(m.model_versions.list(1000))
    latest_version = m.latest_model_version.split("/")[-1]
    if latest_version:
        mv_state = m.model_versions.get(latest_version).state.value
    else:
        mv_state = "N/A"
    current_version = m.current_version.split("/")[-1]
    mv_type = "N/A"
    mv_perf_metric = "N/A"
    mv_perf_metric_val = "N/A"
    if current_version:
        mv = m.model_versions.get(current_version)
        try:
            mv_type = mv.experiment.type.split("/")[1]
        except:
            mv_type = "N/A"
        mv_perf_metric = mv.performance_metric
        mv_perf_metric_val = round(mv.performance_metric_val, 4)
    linked_entities = get_linked_entities(m.schema.columns)
    project = m.parent.split("/")[-1]
    if zipped:
        data = [
            m.id,
            m.name,
            m.state,
            project,
            latest_version,
            mv_state,
            current_version,
            mv_type,
            mv_perf_metric,
            mv_perf_metric_val,
            m.latest_batch_prediction,
            linked_entities,
            num_mv,
            num_d,
            num_bpj,
            create_time,
            update_time,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Project",
            "Latest Model Version",
            "Latest Version State",
            "Current Version",
            "Current Version Type",
            "Current Version Performance Metric",
            "Current Version Performance Metric Value",
            "Latest Batch Prediction",
            "Linked Entities",
            "Model Versions",
            "Promotions",
            "Batch Predictions",
            "Created",
            "Updated",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            m.id,
            m.state,
            latest_version,
            mv_state,
            current_version,
            mv_type,
            mv_perf_metric,
            mv_perf_metric_val,
            linked_entities,
            num_mv,
            num_d,
            num_bpj,
            create_time,
            # update_time,
        ]
        headers = [
            "ID",
            "State",
            "Latest Version",
            "Latest Version State",
            "Current Version",
            "Current Version Type",
            "Current Version Perf Metric",
            "Current Version Perf Metric Val",
            "Linked Entities",
            "Model Versions",
            "Promotions",
            "Batch Predictions",
            "Created",
            # "Updated",
        ]
        if all_projects:
            data.insert(0, m.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        model_list(
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def model_list(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List models.

    Filter can include:
        - state  (i.e. state:HEALTHY)
        - latest_mv_state (i.e. latest_mv_state:SUCCEEDED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if all_projects:
        project_snippet = "all accessible projects"
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    data = []
    headers = []
    for m in c.models.list(n, filters=filters, all_projects=all_projects):
        (m_data, headers) = format_model_data(m, all_projects=all_projects)
        data.append(m_data)
    typer.secho(
        "\nFound %s models in %s%s:" % (len(data), project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Get model details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    m = c.models.get(model)
    console = Console()
    if to_json:
        console.print(m.to_dict())
    else:
        data = format_model_data(m, zipped=True)
        typer.secho("\nRetrieving model %s: \n" % (model), fg="blue")
        utils.print_info(data)
        typer.secho("\nModel schema:\n", fg="blue")
        console.print(m.schema.to_dict())


@app.command("create-local-dev-model-version")
@utils.exit_on_error
def create_local_dev(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Create model version for local development."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.models.create_local_dev_model_version(model)
    console = Console()
    if to_json:
        console.print(mv.to_dict())
    else:
        typer.secho("Created model version %s \n" % (mv.name), fg="blue")


@app.command("create-local-dev-batch-prediction")
@utils.exit_on_error
def create_local_dev(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    to_json: bool = typer.Option(
        False, "--json", help="Show full JSON representation."
    ),
):
    """Create model version for local development."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv = c.models.create_local_dev_model_version(model)
    console = Console()
    if to_json:
        console.print(mv.to_dict())
    else:
        typer.secho("Created model version %s \n" % (mv.name), fg="blue")


@app.command("train")
@utils.exit_on_error
def train(
    model: str = typer.Argument(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    wait: bool = typer.Option(False, "--wait", help="Wait for training to finish."),
    debug: bool = typer.Option(False, "--debug", help="Print full name of model."),
    all: bool = typer.Option(False, "--all", help="Train all models."),
    state: ModelVersionState = typer.Option(
        None, help="Only train models with a certain state."
    ),
    include_estimate: bool = typer.Option(
        False, "--include-estimate", help="Include credit usage estimate for resources."
    ),
):
    """Train model."""
    project, environment = utils.get_project_and_environment(project, environment)
    p_slug = utils.get_environment_name(project, environment)
    c = Client(project=p_slug)
    if all or state is not None:
        model_list = []
        if all == True:
            model_list = c.models.list_all().to_list()
        else:
            for m in c.models.list_all():
                try:
                    mv = c.model_versions.get(m.latest_model_version)
                    if mv.state.value == state:
                        model_list.append(m)
                except:
                    continue

        credits = {}
        if len(model_list) > 0:
            endpoint = utils.get_endpoint()
            for model in model_list:
                try:
                    mv = model.train()
                    typer.secho(
                        "Started training model version %s for model %s.\n"
                        % (mv.id, model.id),
                        fg="green",
                    )
                    typer.secho("You can access the model version at: ", fg="white")
                    typer.secho(
                        f"  {endpoint}/projects/{p_slug}/model/{model.id}/versions/{mv.id}\n",
                        fg="blue",
                    )
                    if include_estimate:
                        credits_used = utils.get_model_credit_estimate(
                            model.name, project, environment
                        )
                        credits["%s %s" % (model.id, "TRAIN")] = credits_used
                except Exception as e:
                    typer.secho(
                        "Failed to train model %s: %s" % (model.id, str(e)), fg="red"
                    )
                    continue

            if include_estimate and len(credits) > 0:
                utils.print_credit_estimate_summary(credits)

        else:
            typer.secho("No models found to train.")

    elif model is not None:
        m = c.models.get(model)
        if include_estimate:
            credits_used = utils.get_model_credit_estimate(m.name, project, environment)
            typer.secho(
                "Estimated Credit Usage for Model: %s\n" % credits_used, fg="white"
            )

        mv = m.train()
        if wait:
            mv.wait()
            typer.secho(
                "Successfully trained model version %s for model %s.\n" % (m.id, mv.id),
                fg="green",
            )
        else:
            typer.secho(
                "Started training model version %s for model %s." % (m.id, mv.id),
                fg="green",
            )
        endpoint = utils.get_endpoint()
        typer.secho("You can access the model version at: ", fg="white")
        typer.secho(
            f"  {endpoint}/projects/{p_slug}/model/{m.id}/versions/{mv.id}\n",
            fg="blue",
        )
        if debug:
            typer.echo("DEBUG: Full name = %s." % mv.name)
    else:
        typer.secho(
            "Error: You must either provide model or one of --all or --state.", fg="red"
        )


@app.command("delete")
@utils.exit_on_error
def delete(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete model."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    if not force:
        force = typer.confirm(
            "Are you sure you want to delete model %s in project %s, environment %s?"
            % (model, project, environment)
        )
    if force:
        m = c.models.get(model)
        m.delete()
        typer.secho("Successfully deleted model %s." % m.id, fg="green")
    else:
        typer.secho("Cancelled model deletion!", fg="blue")


# dev-2222: removing until we have a better way to download data
# @app.command("get-training-data")
# @utils.exit_on_error
# def get_training_data(
#     model: str = typer.Argument(..., help="Model ID."),
#     project: str = typer.Option(None, help="Project ID."),
#     environment: str = typer.Option(None, "--env", help="Environment ID."),
#     path: str = typer.Option(..., help="Path to save training data CSV into."),
#     num_records: int = typer.Option(
#         None, "--num-records", "-n", help="Number of records to retrieve."
#     ),
#     normalized: bool = typer.Option(
#         False,
#         "--normalized",
#         help="Only return base model data (normalized).",
#         hidden=True,
#     ),
#     fetch_metadata: bool = typer.Option(
#         False, "--fetch-metadata", help="Fetch metadata fields ", hidden=True
#     ),
#     latest: bool = typer.Option(
#         False, "--latest", help="Return latest record for each ID.", hidden=True
#     ),
# ):
#     """Get training data for model."""
#     project, environment = utils.get_project_and_environment(project, environment)
#     c = Client(project=utils.get_environment_name(project, environment))

#     typer.secho(
#         "Getting training data for model %s in project %s, environment %s."
#         % (model, project, environment),
#         fg="blue",
#     )
#     try:
#         m = c.models.get(model)
#         (df, _) = m.get_training_data(
#             n=num_records, normalized=normalized, metadata=fetch_metadata, latest=latest
#         )
#         if path[-4:] != ".csv":
#             path = "%s/%s.csv" % (path, model)
#         if df is not None and df.shape[0] > 0:
#             df.to_csv(path, index=False)
#             typer.secho(
#                 "Saved training data for model %s to %s." % (model, path), fg="green"
#             )
#             typer.echo(df.head())
#         else:
#             typer.secho(
#                 "Failed to download data for model %s: Returned no results." % model,
#                 fg="red",
#             )
#     except Exception as e:
#         typer.secho(
#             "Failed to get training data for model %s in project %s environment %s: %s"
#             % (model, project, environment, str(e)),
#             fg="red",
#         )


@app.command("diff")
@utils.exit_on_error
def diff(
    right_yaml: str = typer.Option(None, "--ryaml", help="YAML location."),
    right_project: str = typer.Option(None, "--rproj", help="Right project."),
    right_environment: str = typer.Option(None, "--renv", help="Environment ID."),
    right_model: str = typer.Option(None, "--rmodel", help="Right model."),
    left_yaml: str = typer.Option(None, "--lyaml", help="YAML location."),
    left_project: str = typer.Option(None, "--lproj", help="Left project."),
    left_environment: str = typer.Option(None, "--lenv", help="Environment ID."),
    left_model: str = typer.Option(None, "--lmodel", help="Left model"),
):
    """
    Diff model configuration.

    Can either provide a local YAML file or a model ID already
    contained in Continual. To use local YAML files, use the
    *_yaml options. Otherwise use *_project and *_model options.
    """
    if not ((right_yaml or right_model) and (left_yaml or left_model)):
        typer.secho(
            "Error: You must either provide 2 model references or 2 YAML files to diff models."
        )
    else:
        feature_sets.diff(
            right_yaml,
            right_project,
            right_environment,
            right_model,
            left_yaml,
            left_project,
            left_environment,
            left_model,
            models=True,
        )


@app.command("validate")
@utils.exit_on_error
def validate_yaml(
    path: Path = typer.Argument(..., help="Full path of YAML file."),
):
    """Validate model YAML file."""
    feature_sets.validate_yaml(path)


@app.command("export")
@utils.exit_on_error
def export_yaml(
    model: str = typer.Argument(..., help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    path: str = typer.Option(
        "print",
        help="Path to save YAML (local). Default = print to console.",
        show_default=False,
    ),
):
    """Export YAML file for feature set."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    m = c.models.get(model)
    m_text = yaml.load(m.schema_text, Loader=yaml.FullLoader)
    yaml_text = utils.format_yaml(m_text)
    if path == "print":
        typer.secho(
            "Successfully extracted YAML file from feature set %s." % model,
            fg="green",
        )
        print(yaml_text)
    else:
        utils.save_yamls([(model, yaml_text)], path)
        typer.secho(
            "Successfully extracted YAML file from feature set %s. Location is %s."
            % (model, path),
            fg="green",
        )


@app.command("compare-across-project")
@utils.exit_on_error
def compare_across_project(
    project: str = typer.Argument(None, help="Project ID.", show_default=False),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    models: str = typer.Option(
        None, help="Model IDs, comma separated list", show_default=False
    ),
    performance_metric: utils.PerformanceMetric = typer.Option(
        None, help="Performance Metric to use for comparison."
    ),
    split_type: utils.SplitType = typer.Option(
        utils.SplitType.TEST, help="Performance Metric to use for comparison."
    ),
    stacked: bool = typer.Option(
        True, "--report", help="Print information as a report."
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """
    Compare models in a project or environment.
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    mv_list = []
    try:
        if models is not None:
            ms = models.split(",")
            for model in ms:
                m = c.models.get(model)
                try:
                    model_version = m.model_versions.get(m.current_version)
                except:
                    model_version = m.model_versions.get(m.latest_model_version)
                mv_list.append(model_version)
        else:
            for m in c.models.list_all():
                try:
                    model_version = m.model_versions.get(m.current_version)
                except:
                    model_version = m.model_versions.get(m.latest_model_version)
                mv_list.append(model_version)

        data = []
        metrics = {}
        fi = {}

        typer.secho("Comparing models in project %s:" % (project))
        for mv in mv_list:
            try:
                model_versions.get_model_info(
                    project,
                    environment,
                    mv.id,
                    performance_metric,
                    split_type,
                    data,
                    metrics,
                    fi,
                    include_model=True,
                )
            except:
                print("failure!")
                continue

        if len(data) > 1:
            model_versions.print_mv_dashboard(
                data,
                metrics,
                fi,
                split_type,
                utils.get_style(style),
                include_project=False,
                include_model=True,
                stacked=stacked,
            )
        else:
            typer.secho(
                "Failed to find two or more models that have a successful current or latest model version in project %s, environment %s."
                % (project, environment),
                fg="red",
            )

    except Exception as e:
        typer.secho(
            "Failed to compare models in project %s: %s" % (project, str(e)), fg="red"
        )


@app.command("compare")
@utils.exit_on_error
def compare(
    right_model: str = typer.Option(
        ..., "--right_model", "-rmodel", help="Model ID.", show_default=False
    ),
    right_project: str = typer.Option(
        None, "--right_project", "-rproj", help="Project ID.", show_default=False
    ),
    right_environment: str = typer.Option(
        None, "--right_environment", "-renv", help="Environment ID."
    ),
    left_model: str = typer.Option(
        ..., "--left_model", "-lmodel", help="Model ID.", show_default=False
    ),
    left_project: str = typer.Option(
        None, "--left_project", "-lproj", help="Project ID.", show_default=False
    ),
    left_environment: str = typer.Option(
        None, "--left_environment", "-lenv", help="Environment ID."
    ),
    performance_metric: utils.PerformanceMetric = typer.Option(
        None, help="Performance Metric to use for comparison."
    ),
    split_type: utils.SplitType = typer.Option(
        utils.SplitType.TEST, help="Performance Metric to use for comparison."
    ),
    do_diff: bool = typer.Option(True, help="Diff model schemas."),
    stacked: bool = typer.Option(
        True, "--report", help="Print information as a report."
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """Compare two model versions across projects or environments."""
    data = []
    metrics = {}
    fi = {}

    right_project, right_environment = utils.get_project_and_environment(
        right_project, right_environment
    )
    rc = Client(project=utils.get_environment_name(right_project, right_environment))
    try:
        right_mv_id = rc.models.get(right_model).current_version
    except:
        right_mv_id = rc.models.get(right_model).latest_model_version

    left_project, left_environment = utils.get_project_and_environment(
        left_project, left_environment
    )
    c = Client(project=utils.get_environment_name(left_project, left_environment))
    try:
        left_mv_id = rc.models.get(left_model).current_version
    except:
        left_mv_id = rc.models.get(left_model).latest_model_version

    model_versions.get_model_info(
        right_project,
        right_environment,
        right_mv_id,
        performance_metric,
        split_type,
        data,
        metrics,
        fi,
        include_project=True,
        include_model=True,
    )
    model_versions.get_model_info(
        left_project,
        left_environment,
        left_mv_id,
        performance_metric,
        split_type,
        data,
        metrics,
        fi,
        include_project=True,
        include_model=True,
    )

    model_versions.print_mv_dashboard(
        data,
        metrics,
        fi,
        split_type,
        utils.get_style(style),
        include_project=True,
        include_model=True,
        stacked=stacked,
    )

    if do_diff:
        diff(
            None,
            right_project,
            right_environment,
            right_model,
            None,
            left_project,
            left_environment,
            left_model,
        )



from jinja2 import Template
from continual.python.sdk.query import Query
import logging
from continual.rpc.management.v1 import management_types_pb2
from google.protobuf.json_format import MessageToDict
from continual.python.utils.utils import (
    split_name,
)

TypemapString = {
    "NUMBER": "float",
    "FIXED": "float",
    "REAL": "float",
    "TEXT": "text",
    "BOOLEAN": "boolean",
    "CATEGORICAL": "text",
    "NUMBER_LIST": "[] float",
    "TEXT_LIST": "[] text",
    "TIMESTAMP": "timestamptz",
}

"""
    left_id, left_ts - right_id, right_ts - Condition
    1    ✅      ✅        ✅         ❌         id join
        
    2    ⭐️      ⭐️        ❌         ❌         Invalid
        
    3    ✅      ✅        ❌         ✅         ts join
        
    4    ✅      ✅        ✅         ✅         id and ts join
        
    5    ✅      ❌        ❌         ✅         Invalid
        
    6    ✅      ❌        ✅         ✅         id join
        
    7    ❌      ✅        ✅         ❌         Invalid
        
    8    ❌      ✅        ❌         ✅         ts join
        
    9    ❌      ✅        ✅         ✅         Invalid
        
    10   ❌      ❌        ⭐️         ⭐️         Invalid

    11   ✅      ❌        ✅         ❌         id join
"""
join_decision_dict = {
    (True, True, True, False): "ID",  # 1
    (True, True, False, True): "TS",  # 3
    (True, True, True, True): "IDTS",  # 5
    (True, False, True, True): "ID",  # 6
    (False, True, False, True): "TS",  # 8
    (True, False, True, False): "ID",  # 11
}

# NOTE: changed "<=" tp "=" for opalai temporary workaround
qtemplate = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %}{{ fname }} as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
            , {% if params.Latest and params.TsName and params.PkName %} rank() over (partition by  {{params.PkName}} ORDER BY {{params.TsName}} DESC) as "{{params.TableAlias}}.timerank" 
            {% else %}
             1 as "{{params.TableAlias}}.timerank"
            {% endif %}
        FROM {{ params.EntityTable}}
                {% if params.FilterList %} WHERE {% for filter in params.FilterList %} {% if filter != params.FilterList[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
        {% if params.OrderByClause %} ORDER BY {{ params.OrderByClause }} {% endif %}) AS "{{ params.TableAlias }}"
        {% if params.SubEntity %} ON {% if params.IDJoin %} "{{ params.IDJoin[0] }}" = "{{ params.IDJoin[1] }}" {% endif %} {% if params.TSJoin %} {% if params.IDJoin %} AND {% endif %}"{{ params.TSJoin[0] }}" = ({{ params.TSJoin[1] }} {% endif %} ){% endif %} 
        """
)

qtemplate_no_ranking = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %}{{ fname }} as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
        FROM {{ params.EntityTable}}
                {% if params.FilterList %} WHERE {% for filter in params.FilterList %} {% if filter != params.FilterList[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
        {% if params.OrderByClause %} ORDER BY {{ params.OrderByClause }} {% endif %}) AS "{{ params.TableAlias }}"
        {% if params.SubEntity %} ON {% if params.IDJoin %} "{{ params.IDJoin[0] }}" = "{{ params.IDJoin[1] }}" {% endif %} {% endif %} 
        """
)

qtemplate_mssql = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %} [{{ params.FeatureNames[fname]['column_name'] }}] as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
            , {% if params.Latest and params.TsName %}rank() over (partition by  {{params.PkName}} ORDER BY {{params.TsName}} DESC) as "{{params.Entity}}.timerank" 
            {% else %}
             1 as "{{params.Entity}}.timerank"
            {% endif %}
        FROM {{ params.EntityTable}}
                {% if params.FilterList %} WHERE {% for filter in params.FilterList %} {% if filter != params.FilterList[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
        ) AS {{ params.TableAlias }}
        {% if params.SubEntity %} ON [{{params.MatchId}}] = [{{params.MatchRHS}}] {% if params.RootVisibilityTime and params.TsName %} AND [{{params.QualifiedTs}}] <= [{{params.RootVisibilityTime}}] {% endif %} {% endif %} 
        """
)

qtemplate_mssql_no_ranking = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %} [{{ params.FeatureNames[fname]['column_name'] }}] as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
        FROM {{ params.EntityTable}}
                {% if params.FilterList %} WHERE {% for filter in params.FilterList %} {% if filter != params.FilterList[loop.index0] %} AND {% endif %} {{ filter}} {% endfor %} {% endif %}
        ) AS {{ params.TableAlias }}
        {% if params.SubEntity %} ON [{{params.MatchId}}] = [{{params.MatchRHS}}] {% endif %} 
        """
)

ranktemplate = Template(
    """
    SELECT *
    FROM (SELECT *,
          rank() over (partition by "{{params.PkName}}" {% if not params.Latest and params.TsName %} , "{{params.TsName}}" {% endif %} ORDER BY "{{params.TableAlias}}.timerank" ASC
            {% for subentity in params.SubEntities %} , "{{subentity}}.timerank" ASC {% endfor %}) AS finalrank
          FROM {{params.EntityQuery}}) AS final
    WHERE finalrank = 1
    ORDER BY "{{params.PkName}}" ASC
    {% if params.Limit %} LIMIT {{params.Limit}} {% endif %}
    """
)

ranktemplate_norank = Template(
    """
    SELECT * FROM {{params.EntityQuery}}
    ORDER BY "{{params.PkName}}" ASC
    {% if params.Limit %} LIMIT {{params.Limit}} {% endif %}
    """
)

ranktemplate_mssql = Template(
    """
    SELECT *
    FROM (SELECT *,
          rank() over (partition by [{{params.PkName}}] {% if not params.Latest and params.TsName %} , [{{params.Entity}}:{{params.TsName}}] {% endif %} ORDER BY [{{params.TableAlias}}.timerank] ASC
            {% for subentity in params.SubEntities %} , [{{subentity}}.timerank] ASC {% endfor %}) AS finalrank
          FROM {{params.EntityQuery}}) AS final
    WHERE finalrank = 1
    {% if params.Limit %} LIMIT {{params.Limit}} {% endif %}
    """
)

ranktemplate_mssql_norank = Template(
    """
    SELECT * FROM {{params.EntityQuery}}
    {% if params.Limit %} LIMIT {{params.Limit}} {% endif %}
    """
)

timestamp_template = Template(
    """
    SELECT max({{params.EntityTable}}.{{params.TsName}}) 
    FROM {{params.EntityTable}}
    WHERE {{params.EntityTable}}.{{params.PkName}}="{{params.RootEntityID}}" and {{params.EntityTable}}.{{params.TsName}}<= "{{params.RootVisibilityTime}}"
"""
)

stats_template_snowflake = Template(
    """
 {% for elem in params.AllFeatures %} {% if loop.index > 1 %} UNION {% endif %}
	SELECT {% if elem.Alias %} '{{elem.Alias}}' {% else %} '{{elem.name}}' {% endif %}as "column_name",
{% if (elem.type == "NUMBER" or elem.type == "REAL" or elem.type == "FIXED") %}
	COALESCE(max(allentities.{{elem.name}}), 0.0) as "max", COALESCE(min(allentities.{{elem.name}}), 0.0) as "min",
	COALESCE(avg(allentities.{{elem.name}}), 0.0) as "mean", COALESCE(stddev_pop(TO_DOUBLE(allentities.{{elem.name}})), 0.0) as "sd",
	0 as "min_length", 0 as "max_length", 0 as "distinct_values", NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% elif (elem.type == "TIMESTAMP" or elem.type == "TIMESTAMP_LTZ" or elem.type == "TIMESTAMP_TZ" or elem.type == "TIMESTAMP_NTZ" or elem.type == "DATE") %}
	0.0 as "max", 0.0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values", 
    max(allentities.{{elem.name}})::timestamp as "latest", min(allentities.{{elem.name}})::timestamp as "earliest",
{% elif elem.type == "TEXT" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", COALESCE(min(length(allentities.{{elem.name}})), 0) as "min_length",
	COALESCE(max(length(allentities.{{elem.name}})), 0) as "max_length", COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values",
	NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% elif elem.type == "CATEGORICAL" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd",
	COALESCE(min(length(allentities.{{elem.name}})), 0) as "min_length", COALESCE(max(length(allentities.{{elem.name}})), 0) as "max_length",
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values", NULL::timestamp  as "latest",
	NULL::timestamp  as "earliest",
{% elif elem.type == "BOOLEAN" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length",
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values", NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% else %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values",
	NULL::timestamp  as "latest", NULL::timestamp  as "earliest",{% endif %}
	count(*) - count(allentities.{{elem.name}})  as "null_count",
	count(allentities.{{elem.name}}) as "count",
	CURRENT_TIMESTAMP as "computed_at", '{{elem.ctype}}' as "type"
    FROM  {{params.FullName}} as allentities
	{% endfor %}
    """
)

stats_template = Template(
    """
 {% for elem in params.AllFeatures %} {% if loop.index > 1 %} UNION {% endif %}
	SELECT {% if elem.Alias %} '{{elem.Alias}}' {% else %} '{{elem.name}}' {% endif %}as "column_name",
{% if elem.type == "NUMBER" %}
	COALESCE(max(allentities.{{elem.name}}), 0.0) as "max", COALESCE(min(allentities.{{elem.name}}), 0.0) as "min",
	COALESCE(avg(allentities.{{elem.name}}), 0.0) as "mean", COALESCE(stddev_pop(allentities.{{elem.name}}), 0.0) as "sd",
	0 as "min_length", 0 as "max_length", 0 as "distinct_values", NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% elif elem.type == "TIMESTAMP" %}
	0.0 as "max", 0.0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values", 
    max(allentities.{{elem.name}})::timestamp as "latest", min(allentities.{{elem.name}})::timestamp as "earliest",
{% elif elem.type == "TEXT" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", COALESCE(min(length(allentities.{{elem.name}})), 0) as "min_length",
	COALESCE(max(length(allentities.{{elem.name}})), 0) as "max_length", COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values",
	NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% elif elem.type == "CATEGORICAL" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd",
	COALESCE(min(length(allentities.{{elem.name}})), 0) as "min_length", COALESCE(max(length(allentities.{{elem.name}})), 0) as "max_length",
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values", NULL::timestamp  as "latest",
	NULL::timestamp  as "earliest",
{% elif elem.type == "BOOLEAN" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length",
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as "distinct_values", NULL::timestamp  as "latest", NULL::timestamp  as "earliest",
{% else %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values",
	NULL::timestamp  as "latest", NULL::timestamp  as "earliest",{% endif %}
	count(*) - count(allentities.{{elem.name}})  as "null_count",
	count(allentities.{{elem.name}}) as "count",
	CURRENT_TIMESTAMP as "computed_at", '{{elem.ctype}}' as "type"
    FROM  {{params.FullName}} as allentities
	{% endfor %}
    """
)

stats_template_databricks = Template(
    """
 {% for elem in params.AllFeatures %} {% if loop.index > 1 %} UNION {% endif %}
	SELECT {% if elem.Alias %} '{{elem.Alias}}' {% else %} '{{elem.name}}' {% endif %}as column_name,
{% if elem.type == "NUMBER" %}
	COALESCE(max(allentities.{{elem.name}}), 0.0) as max, COALESCE(min(allentities.{{elem.name}}), 0.0) as min,
	COALESCE(avg(allentities.{{elem.name}}), 0.0) as mean, COALESCE(stddev_pop(allentities.{{elem.name}}), 0.0) as sd,
	0 as min_length, 0 as max_length, 0 as distinct_values, NULL::timestamp  as latest, NULL::timestamp  as earliest,
{% elif elem.type == "TIMESTAMP" %}
	0.0 as max, 0.0 as min, 0.0 as mean, 0.0 as sd, 0 as min_length, 0 as max_length, 0 as distinct_values, 
    max(allentities.{{elem.name}})::timestamp as latest, min(allentities.{{elem.name}})::timestamp as earliest,
{% elif elem.type == "TEXT" %}
	0 as max, 0 as min, 0.0 as mean, 0.0 as sd, COALESCE(min(length(allentities.{{elem.name}})), 0) as min_length,
	COALESCE(max(length(allentities.{{elem.name}})), 0) as max_length, COALESCE(count(distinct(allentities.{{elem.name}})), 0) as distinct_values,
	NULL::timestamp  as latest, NULL::timestamp  as earliest,
{% elif elem.type == "CATEGORICAL" %}
	0 as max, 0 as min, 0.0 as mean, 0.0 as sd,
	COALESCE(min(length(allentities.{{elem.name}})), 0) as min_length, COALESCE(max(length(allentities.{{elem.name}})), 0) as max_length,
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as distinct_values, NULL::timestamp  as latest,
	NULL::timestamp  as earliest,
{% elif elem.type == "BOOLEAN" %}
	0 as max, 0 as min, 0.0 as mean, 0.0 as sd, 0 as min_length, 0 as max_length,
	COALESCE(count(distinct(allentities.{{elem.name}})), 0) as distinct_values, NULL::timestamp  as latest, NULL::timestamp  as earliest,
{% else %}
	0 as max, 0 as min, 0.0 as mean, 0.0 as sd, 0 as min_length, 0 as max_length, 0 as distinct_values,
	NULL::timestamp  as latest, NULL::timestamp  as earliest,{% endif %}
	count(*) - count(allentities.{{elem.name}})  as null_count,
	count(allentities.{{elem.name}}) as count,
	CURRENT_TIMESTAMP as computed_at, '{{elem.ctype}}' as type
    FROM  {{params.FullName}} as allentities
	{% endfor %}
    """
)

stats_template_mssql = Template(
    """
 {% for elem in params.AllFeatures %} {% if loop.index > 1 %} UNION {% endif %}
	SELECT {% if elem.Alias %} '{{elem.Alias}}' {% else %} '{{elem.name}}' {% endif %}as "column_name",
{% if elem.type == "NUMBER" %}
	COALESCE(max([{{elem.name}}]), 0.0) as "max", COALESCE(min([{{elem.name}}]), 0.0) as "min",
	COALESCE(avg([{{elem.name}}]), 0.0) as "mean", COALESCE(stddev_pop([{{elem.name}}]), 0.0) as "sd",
	0 as "min_length", 0 as "max_length", 0 as "distinct_values", CAST(NULL as datetime)  as "latest", CAST(NULL as datetime)  as "earliest",
{% elif elem.type == "TIMESTAMP" %}
	0.0 as "max", 0.0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values", 
    max([{{elem.name}}])::timestamp as "latest", min([{{elem.name}}])::timestamp as "earliest",
{% elif elem.type == "TEXT" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", COALESCE(min(length([{{elem.name}}])), 0) as "min_length",
	COALESCE(max(length([{{elem.name}}])), 0) as "max_length", COALESCE(count(distinct([{{elem.name}}])), 0) as "distinct_values",
	CAST(NULL as datetime)  as "latest", CAST(NULL as datetime)  as "earliest",
{% elif elem.type == "CATEGORICAL" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd",
	COALESCE(min(length([{{elem.name}}])), 0) as "min_length", COALESCE(max(length([{{elem.name}}])), 0) as "max_length",
	COALESCE(count(distinct([{{elem.name}}])), 0) as "distinct_values", CAST(NULL as datetime)  as "latest",
	CAST(NULL as datetime)  as "earliest",
{% elif elem.type == "BOOLEAN" %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length",
	COALESCE(count(distinct([{{elem.name}}])), 0) as "distinct_values", CAST(NULL as datetime)  as "latest", CAST(NULL as datetime)  as "earliest",
{% else %}
	0 as "max", 0 as "min", 0.0 as "mean", 0.0 as "sd", 0 as "min_length", 0 as "max_length", 0 as "distinct_values",
	CAST(NULL as datetime)  as "latest",CAST(NULL as datetime)  as "earliest",{% endif %}
	count(*) - count([{{elem.name}}])  as "null_count",
	count([{{elem.name}}]) as "count",
	CURRENT_TIMESTAMP as "computed_at", '{{elem.ctype}}' as "type"
    FROM  {{params.FullName}} as allentities
	{% endfor %}
    """
)

serving_template_postgres = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT
        {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %}{{ fname }} as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
        {% for oride in params.Overrides %} , CAST({{oride["data"]}} as {{oride["type"]}}) as "{{oride["qualified_name"]}}" {% endfor %}
        {% if params.Latest and params.TsName %}, rank() over (partition by {{params.PkName}} ORDER BY {{params.TsName}} DESC) as "{{params.Entity}}.timerank"
        {% else %}
        , 1 as  "{{params.Entity}}.timerank"
        {% endif %}
        FROM {{params.FullName}}
        {% if params.FilterList %} WHERE {% for filter in params.FilterList%} {% if loop.index > 1 %} AND {% endif %} {{filter}} {% endfor %} {% endif %}
        {% if params.OrderByClause %} ORDER BY {{params.OrderByClause}} {% endif %}) AS "{{params.Entity}}"
    {% if params.SubEntity %} ON "{{params.MatchId}}" = "{{params.MatchRHS}}" {% endif %}
"""
)

serving_template_mssql = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT
        {% for fname in params.FeatureNames %} {% if loop.index > 1 %},{% endif %} [{{ params.FeatureNames[fname]['column_name'] }}] as "{{ params.FeatureNames[fname]['qualified'] }}" {% endfor %}
        {% for oride in params.Overrides %} , CAST({{oride["data"]}} as {{oride["type"]}}) as "{{oride["qualified_name"]}}" {% endfor %}
        {% if params.Latest and params.TsName %}, rank() over (partition by {{params.PkName}} ORDER BY {{params.TsName}} DESC) as "{{params.Entity}}.timerank"
        {% else %}
        , 1 as  "{{params.Entity}}.timerank"
        {% endif %}
        FROM {{params.FullName}}
        {% if params.FilterList %} WHERE {% for filter in params.FilterList%} {% if loop.index > 1 %} AND {% endif %} {{filter}} {% endfor %} {% endif %}
        ) AS {{params.Entity}}
    {% if params.SubEntity %} ON [{{params.MatchId}}] = [{{params.MatchRHS}}] {% endif %}
"""
)

serving_template_noid_postgres = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT
    {% for oride in params.Overrides %} CAST({{oride["data"]}} as {{oride["type"]}}) as "{{oride["qualified_name"]}}", {% endfor %}
    1 as "{{params.Entity}}.timerank"
    ) AS "{{params.Entity}}" {% if params.SubEntity %} ON TRUE {% endif %}
"""
)

serving_template_noid_mssql = Template(
    """
    {% if params.SubEntity %} LEFT JOIN {% endif %}
    (SELECT
    {% for oride in params.Overrides %} CAST({{oride["data"]}} as {{oride["type"]}}) as "{{oride["qualified_name"]}}", {% endfor %}
    1 as "{{params.Entity}}.timerank"
    ) AS {{params.Entity}} {% if params.SubEntity %} ON 1=1 {% endif %}
"""
)
serving_rank_template_postgres = Template(
    """
    SELECT *
        FROM (SELECT *,
            rank() over (partition by {% if params.IDPresent %} "{{params.PkName}}" {% else %} "{{params.Entity}}.timerank" {% endif %} ORDER BY "{{params.Entity}}.timerank" ASC
                {% for subentity in params.SubEntities %} , "{{subentity}}.timerank" ASC  {% endfor %}) AS finalrank
            FROM {{params.EntityQuery}}) AS final
        WHERE finalrank = 1 {% for idfield in params.IDList %} AND "{{idfield}}" IS NOT NULL {% endfor %}
        {% if params.IDPresent %} ORDER BY "{{params.PkName}}" ASC {% endif %}
"""
)

serving_rank_template_mssql = Template(
    """
    SELECT *
        FROM (SELECT *,
            rank() over (partition by {% if params.IDPresent %} "{{params.PkName}}" {% else %} "{{params.Entity}}.timerank" {% endif %} ORDER BY [{{params.Entity}}.timerank] ASC
                {% for subentity in params.SubEntities %} , "{{subentity}}.timerank" ASC  {% endfor %}) AS finalrank
            FROM {{params.EntityQuery}}) AS final
        WHERE finalrank = 1 {% for idfield in params.IDList %} AND [{{idfield}}] IS NOT NULL {% endfor %}
"""
)

incremental_prediction_template = Template(
    """
    SELECT * from (
    SELECT A.*
    FROM ({{params.TrainingQuery}}) as A
    LEFT JOIN
    ( 
        SELECT id, count(*) pred_count
        from {{params.DbSchema}}.model_{{params.Entity}}_{{params.ModelID}}_preds
        GROUP BY id
    ) as B
    ON "{{params.Entity}}:{{params.PkName}}" = B.id
    WHERE pred_count is NULL
    ) as C
"""
)


def get_index(qdict):
    for field in qdict["fields"]:
        if ("type" in field and field["type"] == "INDEX") or (
            "index" in field and field["index"]
        ):
            return field["name"], field["qualified_name"]
    return None, None


def get_time_index_field(qdict):
    for field in qdict["fields"]:
        if ("type" in field and field["type"] == "TIME_INDEX") or (
            "time_index" in field and field["time_index"]
        ):
            return field["name"], field["qualified_name"]
    return None, None


def create_time_subquery(table_name, pkName, match_id, tsName, root_ts_name):
    params = {
        "EntityTable": table_name,
        "TsName": tsName,
        "PkName": pkName,
        "RootEntityID": match_id,
        "RootVisibilityTime": root_ts_name,
    }
    return timestamp_template.render(params=params)


def create_serving_query(
    query,
    sub_query,
    feature_store,
    match_id=None,
    root_entity_name=None,
    root_ts_name=None,
    start_time=None,
    end_time=None,
    limit=None,
    parent_id_found=False,
):

    """
    Stage 1 for the join queries for all the relations from the spine
    """
    if isinstance(query, Query):
        qdict = query.to_dict()
    elif isinstance(query, management_types_pb2.Query):
        qdict = MessageToDict(query, preserving_proto_field_name=True)
    else:
        qdict = query

    _, table_name = feature_store.get_data_table_name(qdict["name"])

    full_name = qdict.get("table_name", None)
    if full_name is None:
        full_name = feature_store.db_schema + "." + table_name

    entity_name, name_map = split_name(qdict["name"])

    idvalue = None
    id_present = False
    filters = []
    pkName, qualified_pk = get_index(qdict)
    tsName, qualified_ts = get_time_index_field(qdict)

    # If this object is a Model and not a featureset, then
    # the column names are not qualified.
    if "models" in name_map:
        qualified_pk = pkName
        qualified_ts = tsName

    sub_entity_names = []
    id_field_list = []

    params = {
        "Entity": entity_name,
        "EntityTable": full_name,
        "Overrides": [],
        "FeatureNames": {},
        "SubEntity": sub_query,
        "PkName": pkName,
        "TsName": tsName,
        "QualifiedTsName": qualified_ts,
        "FullName": full_name,
    }

    logging.debug(qdict["fields"])
    join_query = ""
    for field in qdict["fields"]:
        if "type" in field and field["type"].upper() == "INDEX":
            id_present = True
            if "value" in field:
                idvalue = field["value"]
                if isinstance(idvalue, str):
                    idvalue = "'{}'".format(idvalue)
                if not sub_query:
                    filters.append("{} = {}".format(pkName, idvalue))
                    field_name = table_name + "." + field["name"]
                    params["FeatureNames"][field_name] = {
                        "qualified": field["qualified_name"],
                        "column_name": field["name"],
                    }
                    continue

        if "value" in field and field["value"] is not None:
            value = field["value"]
            if isinstance(value, str):
                value = "'{}'".format(value)

            type_to_use = field["dtype"]
            if field["dtype"] in TypemapString:
                type_to_use = TypemapString[field["dtype"]]
            params["Overrides"].append(
                {
                    "data": value,
                    "qualified_name": field["qualified_name"],
                    "type": type_to_use,
                }
            )
        else:
            field_name = table_name + "." + field["name"]
            params["FeatureNames"][field_name] = {
                "qualified": field["qualified_name"],
                "column_name": field["name"],
            }

        if "relation" in field:
            # If we filed is found with value then the relations have parent id.
            # else if the current featureset ID is set then the parrent id is set.
            # or if any of the parents have id set, then this set parent_id_found.
            pif = (
                ("value" in field and field["value"] is not None)
                or parent_id_found
                or idvalue != None
            )

            sub_entity_query, sub_entity_n, id_flist = create_serving_query(
                field["relation"],
                True,
                feature_store,
                root_entity_name=entity_name,
                root_ts_name=root_ts_name,
                match_id=field["qualified_name"],
                parent_id_found=pif,
            )
            join_query = join_query + sub_entity_query
            sub_entity_names.extend(sub_entity_n)
            id_field_list.extend(id_flist)

    if not sub_query:
        root_ts_name = qualified_ts
        match_id = qualified_pk
        params["OrderByClause"] = pkName

        if tsName and end_time is not None:
            filters.append("{} <= {}".format(tsName, end_time))

        if tsName and start_time is not None:
            filters.append("{} > {}".format(tsName, start_time))

    else:
        params["MatchId"] = match_id
        params["RootEntityName"] = root_entity_name
        params["RootVisibilityTime"] = root_ts_name
        params["MatchRHS"] = qualified_pk
        qualified_ts = root_ts_name

    if "related" in qdict:
        for sub_qdict in qdict["related"]:
            sub_entity_query, sub_entity_n, id_flist = create_serving_query(
                sub_qdict,
                True,
                feature_store,
                root_entity_name=entity_name,
                root_ts_name=root_ts_name,
                match_id=match_id,
                parent_id_found=(parent_id_found or idvalue != None),
            )
            join_query = join_query + sub_entity_query
            sub_entity_names.extend(sub_entity_n)
            id_field_list.extend(id_flist)

    params["FilterList"] = filters

    # Select template based on dialect.
    template = serving_template_postgres
    if feature_store.data_store.type == "sqlserver":
        template = serving_template_mssql

    if sub_query:
        if not parent_id_found and idvalue is None:
            logging.debug("Using no id template")
            template = serving_template_noid_postgres
            if feature_store.data_store.type == "sqlserver":
                template = serving_template_noid_mssql
    elif idvalue is None:
        logging.debug("Using no id template")
        template = serving_template_noid_postgres
        if feature_store.data_store.type == "sqlserver":
            template = serving_template_noid_mssql
    else:
        logging.debug("Using with id template")
        id_field_list.append(qualified_pk)

    ranktemp = serving_rank_template_postgres
    if feature_store.data_store.type == "sqlserver":
        ranktemp = serving_rank_template_mssql

    entity_query = template.render(params=params)
    entity_query = entity_query.strip() + "\n" + join_query

    """
    Stage 2: Create the rank query if this is the spine.
    """
    if not sub_query:
        params = {
            "IDPresent": (id_present and idvalue is not None),
            "EntityQuery": entity_query,
            "Entity": entity_name,
            "SubEntities": sub_entity_names,
            "PkName": pkName,
            "TsName": tsName,
            "IDList": id_field_list,
        }
        if limit is not None:
            params["Limit"] = limit
        entity_query = ranktemp.render(params=params)

    return entity_query, sub_entity_names, id_field_list


def create_query(
    query,
    sub_query,
    match_id=None,
    root_entity_name=None,
    root_ts_name=None,
    latest=True,
    all_instances=True,
    start_time=None,
    end_time=None,
    limit=None,
    feature_store=None,
    incremental_prediction=False,
    level=0,
):

    """
    Stage 1 for the join queries for all the relations from the spine
    """
    if isinstance(query, Query):
        qdict = query.to_dict()
    elif isinstance(query, management_types_pb2.Query):
        qdict = MessageToDict(query, preserving_proto_field_name=True)
    else:
        qdict = query

    level = level + 1
    _, table_name = feature_store.get_data_table_name(qdict["name"])

    full_name = qdict.get("table_name", None)
    if full_name is None:
        full_name = feature_store.db_schema + "." + table_name

    entity_name, name_map = split_name(qdict["name"])

    query_params = {}
    filters = []
    sub_entity_names = []
    pkName, qualified_pk = get_index(qdict)
    tsName, qualified_ts = get_time_index_field(qdict)

    table_alias = f"{entity_name}_{level}"
    # If this object is a Model and not a featureset, then
    # the column names are not qualified.
    if "models" in name_map:
        qualified_pk = pkName
        qualified_ts = tsName

    params = {
        "Entity": entity_name,
        "TableAlias": table_alias,
        "EntityTable": full_name,
        "FeatureNames": {},
        "SubEntity": sub_query,
        "AllInstances": all_instances,
        "Latest": latest,
        "PkName": pkName,
        "TsName": tsName,
        "QualifiedTsName": qualified_ts,
    }

    join_query = ""

    if not sub_query:
        root_ts_name = qualified_ts
        match_id = qualified_pk
        params["OrderByClause"] = pkName

        if tsName and end_time is not None:
            filters.append("{} <= {}".format(tsName, end_time))
            query_params["visibilitytimeend"] = end_time

        if tsName and start_time is not None:
            filters.append("{} > {}".format(tsName, start_time))
            query_params["visibilitytimestart"] = start_time

    else:
        sub_entity_names.append(table_alias)
        qualified_ts = root_ts_name

        time_sub_query = None
        if tsName and root_ts_name:
            time_sub_query = create_time_subquery(
                full_name, pkName, match_id, tsName, root_ts_name
            )

        des = join_decision_dict.get(
            (
                match_id is not None,
                root_ts_name is not None,
                qualified_pk is not None,
                time_sub_query is not None,
            ),
            "invalid",
        )
        if des == "ID":
            params["IDJoin"] = (match_id, qualified_pk)
        elif des == "TS":
            params["TSJoin"] = (root_ts_name, time_sub_query)
        elif des == "IDTS":
            params["IDJoin"] = (match_id, qualified_pk)
            params["TSJoin"] = (root_ts_name, time_sub_query)
        else:
            raise Exception("Invalid Join")

    if "related" in qdict:
        for sub_qdict in qdict["related"]:

            (sub_entity_query, sub_q_params, sub_entity_n,) = create_query(
                sub_qdict,
                True,
                root_entity_name=entity_name,
                root_ts_name=root_ts_name,
                match_id=match_id,
                latest=latest,
                all_instances=all_instances,
                feature_store=feature_store,
                level=level,
            )
            join_query = join_query + sub_entity_query
            query_params.update(sub_q_params)
            sub_entity_names.extend(sub_entity_n)

    # Set the root_ts_name param so the related featureset query
    # can get them.
    # if not sub_query:
    #    root_ts_name = tsName

    model_id = None
    for field in qdict["fields"]:
        field_name = full_name + "." + field["name"]
        params["FeatureNames"][field_name] = {
            "qualified": field["qualified_name"],
            "column_name": field["name"],
        }
        if "model" in field:
            if not all_instances:
                filters.append(field["name"] + " is NOT NULL")

            model_id = field["name"]

        if "relation" in field:
            (sub_entity_query, sub_q_params, sub_entity_n,) = create_query(
                field["relation"],
                True,
                root_entity_name=entity_name,
                root_ts_name=root_ts_name,
                match_id=field["qualified_name"],
                latest=latest,
                all_instances=all_instances,
                feature_store=feature_store,
                level=level,
            )
            join_query = join_query + sub_entity_query
            query_params.update(sub_q_params)
            sub_entity_names.extend(sub_entity_n)

    params["FilterList"] = filters

    # Select template based on dialect.
    template = None
    ranktemp = None
    if (
        feature_store is not None
        and feature_store.data_store is not None
        and feature_store.data_store.type == "sqlserver"
    ):
        if root_ts_name == None:
            template = qtemplate_mssql_no_ranking
            ranktemp = ranktemplate_mssql_norank
        else:
            template = qtemplate_mssql
            ranktemp = ranktemplate_mssql
    else:
        if root_ts_name == None:
            template = qtemplate_no_ranking
            ranktemp = ranktemplate_norank
        else:
            template = qtemplate
            ranktemp = ranktemplate

    entity_query = template.render(params=params)
    entity_query = entity_query.strip() + "\n" + join_query

    """
	Stage 2: Create the rank query if this is the spine.
	"""
    if not sub_query:
        params = {
            "EntityQuery": entity_query,
            "Entity": entity_name,
            "TableAlias": table_alias,
            "Latest": latest,
            "SubEntities": sub_entity_names,
            "PkName": pkName,
            "TsName": tsName,
        }
        if limit is not None:
            params["Limit"] = limit
            query_params["limit"] = limit
        entity_query = ranktemp.render(params=params)

        if incremental_prediction:
            params = {
                "TrainingQuery": entity_query,
                "ModelID": model_id,
                "Entity": entity_name,
                "PkName": pkName,
                "DbSchema": feature_store.db_schema,
            }
            entity_query = incremental_prediction_template.render(params=params)

    return entity_query, query_params, sub_entity_names



import os
from datetime import datetime
import json
from textwrap import dedent
import time
import tempfile
import logging
import random
from typing import List
from jinja2 import Template
import uuid
import requests
import traceback
from functools import wraps
import sqlalchemy as db
from sqlalchemy.exc import (
    IntegrityError,
    NoSuchTableError,
    ProgrammingError,
    OperationalError,
)
from sqlalchemy.sql.schema import MetaData
from continual.python.sdk.models import Model

from continual.services.dataingest.gcs_source import FeatureSourceCloudStore

from .featurestore import FeatureStore, Typemap
from .prediction_store import PredictionStore, chunk_it

import pandas as pd

from continual.python.sdk.exceptions import InvalidArgumentError
from continual.python.utils.utils import (
    should_enable_databricks_registration,
)

from continual.rpc.management.v1 import management_types_pb2

from .query_templates import stats_template_databricks

from .dialects import DatabricksDialect  # don't remove

logging.getLogger("pyhive").setLevel(logging.WARNING)
logging.getLogger("databricks.sql.client").setLevel(logging.WARNING)


def handle_databricks_exception(func):
    """Decorator that parses Databricks exceptions raised through SQLAlchemy.

    Args:
        func: Function to wrap.
    Returns:
        A function that will raise an informative exception.
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OperationalError as e:
            if "databricks.sql.exc.OperationalError" in str(e):
                error_str = str(e)
                for line in error_str.splitlines():
                    if (
                        line.startswith("Error message:")
                        and "org.apache.spark.SparkDriverExecutionException:"
                        not in line
                    ):
                        raise InvalidArgumentError(message=line, details=None)
                for line in error_str.splitlines():
                    if (
                        line.startswith("Caused by:")
                        and "org.apache.spark.SparkDriverExecutionException" not in line
                    ):
                        raise InvalidArgumentError(message=line, details=None)
                raise
        except:
            raise

    return wrapped_func


WORKSPACE_DIR = "/FileStore/Continual"
SUPPORTED_RUNTIMES = set(["10.5.x-cpu-ml-scala2.12"])
SYNC_JOB_TASK_KEY = "sync"

CONTINUAL_USER_AGENT = "Continual/1.0"


class DatabricksFeatureStore(FeatureStore, PredictionStore):

    DATABRICKS_TYPEMAP = {
        "object": "STRING",
        "int64": "INT",
        "bool": "BOOLEAN",
        "float64": "FLOAT",
        "datetime64": "TIMESTAMP",
        "datetime64[ns]": "TIMESTAMP",
        "timedelta": "STRING",
        "timedelta[ns]": "STRING",
        "category": "STRING",
        "FIXED": "FLOAT",
        "REAL": "FLOAT",
        "FLOAT": "FLOAT",
        "TEXT": "STRING",
        "DATE": "DATE",
        "TIMESTAMP": "TIMESTAMP",
        "VARIANT": "STRING",
        "OBJECT": "STRING",
        "ARRAY": "ARRAY<STRING>",
        "BINARY": "BINARY",
        "TIME": "TIMESTAMP",
        "BOOL": "BOOLEAN",
        "NUMBER": "FLOAT",
        "FLOAT64": "FLOAT",
        "INTEGER": "INT",
        "INT64": "INT",
        "SMALLINT": "INT",
        "TINYINT": "INT",
        "BIGINT": "INT",
        "NUMERIC": "FLOAT",
        "DECIMAL": "FLOAT",
        "BIGNUMERIC": "FLOAT",
        "BIGDECIMAL": "FLOAT",
    }

    def __init__(
        self,
        data_store,
        source,
        cfg: dict = None,
        environment: management_types_pb2.Environment = None,
        revive_cluster=False,
    ):
        self.cfg = cfg or dict()
        logging.info(f"databricks config: {self.cfg}")
        self.data_store = data_store
        config = data_store.databricks
        db_url = "databricks+connector://token:{}@{}:{}/default".format(
            config.token, config.host, config.port
        )
        self.is_gcp = ".gcp." in config.host
        self.is_azure = "adb-" in config.host or "azuredatabricks" in config.host
        self.is_aws = not self.is_gcp and not self.is_azure

        self.enable_registration = should_enable_databricks_registration(
            self.data_store
        )

        if revive_cluster:
            self._revive_cluster()

        self.engine = db.create_engine(
            db_url,
            connect_args={
                "http_path": config.http_path,
                "user-agent": CONTINUAL_USER_AGENT,
            },
        )
        self.connection = self.engine.connect()
        self.trans = None
        self.environment = environment

        super(DatabricksFeatureStore, self).__init__()

    def set_db_schema(self, db_schema):
        self.db_schema = db_schema

    @handle_databricks_exception
    def cleanup(self):
        self.connection.execute(
            "DROP DATABASE IF EXISTS {} CASCADE".format(self.db_schema)
        )
        return

    @handle_databricks_exception
    def close(self):
        self.connection.close()

    @handle_databricks_exception
    def create_project_schema(self):
        if self.db_schema == None:
            raise InvalidArgumentError("schema_name is required")
        # try to see if the schema exists first
        try:
            print(f"Attempting to create project schema: {self.db_schema}")
            temp_id = random.randint(0, 10000)
            self.connection.execute(
                "CREATE TABLE IF NOT EXISTS {}.test_permissions_{} (id INTEGER)".format(
                    self.db_schema, temp_id
                )
            )
            self.connection.execute(
                "DROP TABLE IF EXISTS {}.test_permissions_{}".format(
                    self.db_schema, temp_id
                )
            )
        except Exception:
            logging.debug("database does not appear to exist - create")
            self.connection.execute(
                "CREATE DATABASE IF NOT EXISTS {}".format(self.db_schema)
            )
            logging.debug("database created successfully.")

    @handle_databricks_exception
    def create_featureset_view(self, feature_set: management_types_pb2.FeatureSet):
        project, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        query = feature_set.schema.query.strip().rstrip(";")
        view_name = "{}.{}".format(self.db_schema, table_name)
        # self.connection.execute(
        #     "CREATE OR REPLACE VIEW {} AS ({})".format(
        #         view_name, query
        #     )
        # )

        feature_table_url = "/2.0/feature-store/feature-tables/get?name={}".format(
            view_name
        )

        sync_action = "create"
        res = self._rest_api_call(feature_table_url, method="GET")
        logging.info(f"feature table GET: {res}")
        if res.get("error_code") != "RESOURCE_DOES_NOT_EXIST":
            sync_action = "update"

        logging.info("feature_set: {}".format(feature_set))

        start = time.time()
        logging.info(
            "create_featureset_view kicking off sync job for creating databricks feature table ..."
        )

        try:
            self._run_databricks_sync(
                sync_action,
                {"feature_set_name": feature_set.name, "table": view_name},
            )
            res = self._rest_api_call(
                "/2.0/feature-store/feature-tables/update",
                method="PATCH",
                body={
                    "name": view_name,
                    "description": "\n".join(
                        [
                            "# ![](https://assets.website-files.com/608c6fda09a6be56ba6844df/608c6fda09a6befa68684532_Symbol.svg)&nbsp;&nbsp;&nbsp;&nbsp;Declaratively Managed by Continual.ai ",
                            "## Description",
                            feature_set.schema.description
                            or "_{}_".format("No description provided."),
                            "## Documentation",
                            feature_set.schema.documentation
                            or "_{}_".format("No documentation provided."),
                            "## Links",
                            "Feature Set: {}".format(
                                "https://cloud.continual.ai/" + feature_set.name
                            ),
                            "Edit Feature Set: {}".format(
                                "https://cloud.continual.ai/"
                                + feature_set.name
                                + "/edit"
                            ),
                            "## Current YAML",
                            "```yaml",
                            feature_set.schema_text,
                            "```",
                        ]
                    ),
                },
            )
            logging.info("update call to feature set: {}".format(res))
        except Exception as e:
            logging.info(
                "create_featureset_view sync job took {} seconds".format(
                    time.time() - start
                )
            )
            raise e

    @handle_databricks_exception
    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.
        Returns None
        """
        if feature_set.schema.query is not None and feature_set.schema.query != "":
            if self.enable_registration:
                self.create_featureset_view(feature_set)
            else:
                super().create_featureset_view(feature_set)
            return

        _, table_name = self.get_table_name(feature_set)
        table_name = "featureset_{}".format(table_name)
        meta = db.MetaData(schema=self.db_schema)

        self.create_project_schema()

        columns = []

        for f in feature_set.schema.columns:
            columns.append(db.Column(f.name, Typemap[f.type]))

        db.Table(table_name, meta, *columns)

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                meta.create_all(self.engine)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug("Error while creating feature table " + str(count))
                count = count + 1
                if count == 2:
                    done = True

        self.connection.execute(
            "CONVERT TO DELTA {}.{} NO STATISTICS".format(self.db_schema, table_name)
        )

        if not self.enable_registration:
            return

        table_name = "{}.{}".format(self.db_schema, table_name)

        feature_table_url = "/2.0/feature-store/feature-tables/get?name={}".format(
            table_name
        )

        sync_action = "create"
        res = self._rest_api_call(feature_table_url, method="GET")
        if res.get("error_code") != "RESOURCE_DOES_NOT_EXIST":
            sync_action = "update"

        start = time.time()
        logging.info(
            "create_feature_table kicking off sync job for creating databricks feature table ..."
        )

        try:
            self._run_databricks_sync(
                sync_action,
                {"feature_set_name": feature_set.name, "table": table_name},
            )
        except Exception as e:
            logging.info(
                "create_feature_table sync job took {} seconds".format(
                    time.time() - start
                )
            )
            raise e

    @handle_databricks_exception
    def delete_project_schema(self, schema_name):
        self.connection.execute(
            "DROP DATABASE IF EXISTS {} CASCADE".format(schema_name)
        )

    @handle_databricks_exception
    def drop_featureset_view(self, name):
        if not self.enable_registration:
            return super().drop_featureset_view(name)

        project, table_name = self.get_data_table_name(name)
        # self.connection.execute(
        #     "DROP VIEW IF EXISTS {}.{}".format(self.db_schema, table_name)
        # )

        feature_table_name = "{}.{}".format(self.db_schema, table_name)
        feature_table_url = "/2.0/feature-store/feature-tables/get?name={}".format(
            feature_table_name
        )
        res = self._rest_api_call(feature_table_url, method="GET")
        if res.get("error_code") != "RESOURCE_DOES_NOT_EXIST":
            start = time.time()
            logging.info(
                "create_feature_table kicking off sync job for creating databricks feature table ..."
            )

            try:
                self._run_databricks_sync("delete", {"table": feature_table_name})
            except Exception as e:
                logging.info(
                    "create_feature_table sync job took {} seconds".format(
                        time.time() - start
                    )
                )
                raise e
            finally:
                try:
                    self.connection.execute(
                        "DROP TABLE IF EXISTS {}".format(feature_table_name)
                    )
                except Exception as e:
                    logging.error(e)
        else:
            try:
                self.connection.execute(
                    "DROP VIEW IF EXISTS {}".format(feature_table_name)
                )
            except Exception as e:
                logging.error(e)

        return

    def get_type(self) -> str:
        return "databricks"

    def is_hosted(self):
        return False

    @handle_databricks_exception
    def get_database_names(self):
        databases = []
        result_proxy = self.connection.execute("SHOW DATABASES")
        for rowproxy in result_proxy:
            d = {}
            for col, value in rowproxy.items():
                d[col] = value
            databases.append(d["databaseName"])
        return databases

    def get_table_names(self, database, schema):
        tables = []
        result_proxy = self.connection.execute("SHOW TABLES FROM {}".format(database))
        for rowproxy in result_proxy:
            d = {}
            for col, value in rowproxy.items():
                d[col] = value
            tables.append(d["tableName"])
        return tables

    @handle_databricks_exception
    def get_column_names(self, database, schema, table):
        columns = []
        type_map = {}

        result_proxy = self.connection.execute(
            "DESCRIBE TABLE {}.{}".format(database, table)
        )
        for rowproxy in result_proxy:
            d = {}
            for col, value in rowproxy.items():
                d[col] = value

            columns.append(d["col_name"])
            if d["data_type"]:
                type_map[d["col_name"]] = d["data_type"]

        return columns, type_map

    @handle_databricks_exception
    def get_schema_names(self, database):
        # Databricks does not have Schemas, this is equivalent to SHOW DATABASES
        return self.get_database_names()

    @handle_databricks_exception
    def get_stats_query_template(self):
        return stats_template_databricks

    @handle_databricks_exception
    def get_prediction_table(self, model):
        project_id, pred_table = self.get_model_prediction_table_name(model)
        print("Prediction table in " + self.db_schema)
        try:
            tab = db.Table(
                pred_table,
                db.MetaData(schema=self.db_schema),
                autoload=True,
                autoload_with=self.engine,
            )
            return tab
        except NoSuchTableError:
            # Todo perhaps not capture the error here.
            return None

    @handle_databricks_exception
    def _get_existing_ml_clusters(self) -> List[str]:
        cluster_ids = []
        resp = self._rest_api_call("/2.0/clusters/list", method="GET")

        if "clusters" in resp:
            logging.debug(
                f"_get_existing_ml_clusters - all clusters: {resp.get('clusters')}"
            )
            for cluster in resp["clusters"]:
                if (
                    cluster.get("state", "") == "RUNNING"
                    and cluster.get("spark_version", "") in SUPPORTED_RUNTIMES
                    and cluster.get("cluster_id")
                ):
                    cluster_ids.append(cluster["cluster_id"])

        return cluster_ids

    @handle_databricks_exception
    def _get_all_clusters(self) -> List[dict]:
        clusters = []
        resp = self._rest_api_call("/2.0/clusters/list", method="GET")

        if "clusters" in resp:
            for cluster in resp["clusters"]:
                if cluster.get("cluster_id"):
                    clusters.append(cluster)

        return clusters

    @handle_databricks_exception
    def _revive_cluster(self):
        clusters = self._get_all_clusters()
        if any(
            map(
                lambda c: c.get("state") == "RUNNING"
                and c.get("cluster_id") in self.data_store.databricks.http_path,
                clusters or list(),
            )
        ):
            logging.debug("Existing cluster detected. Exiting _revive_cluster ...")
        elif len(clusters) > 0:
            cluster_to_start = clusters[0]
            cluster_id = cluster_to_start["cluster_id"]

            resp = self._rest_api_call(
                "/2.0/clusters/start", body={"cluster_id": cluster_id}, method="POST"
            )

            resp = self._rest_api_call(
                "/2.0/clusters/get", body={"cluster_id": cluster_id}
            )
            logging.info("Waiting for cluster to start ...")
            while resp.get("state") != "RUNNING":
                time.sleep(5)
                resp = self._rest_api_call(
                    "/2.0/clusters/get", body={"cluster_id": cluster_id}
                )
                print(resp.get("state"))

            logging.debug("Successfully revived cluster.")

    @property
    def single_node_cluster_config(self) -> dict:
        node_type = "m4.large"  # aws
        if self.is_azure:
            node_type = "Standard_DS3_v2"
        elif self.is_gcp:
            node_type = "n1-standard-4"

        config = {
            "spark_version": "10.5.x-cpu-ml-scala2.12",
            "node_type_id": node_type,
            "driver_node_type_id": node_type,
            "num_workers": 0,
            "spark_conf": {
                "spark.databricks.cluster.profile": "singleNode",
                "spark.master": "local[*]",
            },
            "custom_tags": {"ResourceClass": "SingleNode"},
            "autotermination_minutes": 10,
        }

        if self.is_aws:
            config.update(
                {
                    "enable_elastic_disk": True,
                    "aws_attributes": {"availability": "ON_DEMAND"},
                }
            )

        return config

    @handle_databricks_exception
    def _run_databricks_sync(self, action: str, args: dict) -> dict:
        template_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "databricks_sync.jinja2"
        )

        context_id = ""
        cluster_id = ""
        rendered_template_path = ""

        try:
            with tempfile.NamedTemporaryFile(delete=False) as temp:
                with open(template_path) as template_file:
                    template = Template(template_file.read())

                    api_key = self.cfg.get("authorization", "")
                    endpoint = (
                        self.cfg.get("env_config", dict())
                        .get("Endpoints", dict())
                        .get("management_external", "https://api.continual.ai")
                    )

                    with open(temp.name, "w") as f:
                        f.write(
                            template.render(
                                action=action,
                                api_key=api_key,
                                endpoint=endpoint,
                                args=args,
                            )
                        )
                    with open(temp.name) as f:
                        logging.info(f.read())
                    rendered_template_path = temp.name

            existing_cluster_ids = self._get_existing_ml_clusters()
            if existing_cluster_ids:
                cluster_id = random.choice(existing_cluster_ids)
                logging.info(
                    "randomly selected cluster w/ ID ({}) for Databricks sync job.".format(
                        cluster_id
                    )
                )
            else:
                logging.info(
                    "no existing cluster found with a supported runtime. dynamically creating ..."
                )
                cluster_id = self._create_databricks_cluster()

            self._install_continual_pypi_library(cluster_id)

            res = self._rest_api_call(
                "/1.2/contexts/create", {"clusterId": cluster_id, "language": "python"}
            )
            if res.get("error") or not res.get("id"):
                raise ValueError("could not start execution context")

            context_id = res.get("id")

            command_id = ""
            res = self._rest_api_call(
                "/1.2/commands/execute",
                files={
                    "command": (None, open(rendered_template_path, "rb")),
                    "clusterId": cluster_id,
                    "contextId": context_id,
                    "language": "python",
                },
            )

            if res.get("error") or not res.get("id"):
                raise ValueError("could not execute command")

            command_id = res.get("id")

            start = time.time()
            state = ""
            while state not in ("CANCELLED", "ERROR", "FINISHED"):
                res = self._rest_api_call(
                    "/1.2/commands/status",
                    params={
                        "clusterId": cluster_id,
                        "contextId": context_id,
                        "commandId": command_id,
                    },
                    method="GET",
                )
                if res.get("error"):
                    break
                logging.info(f"Sync res: {res}")
                state = res.get("status", "").upper()

                logging.info(
                    "Elapsed time so far: {} seconds ...".format(time.time() - start)
                )
                time.sleep(10)

            logging.info(f"Final databricks response: {res}")
            logging.info(
                "Databricks sync task returned state ({}) in {} seconds ...".format(
                    state, time.time() - start
                )
            )

            if res.get("error"):
                raise ValueError(res["error"])
            elif res.get("results", dict()).get("resultType", "") == "error":
                results = res.get("results", dict())
                summary = results.get("summary", "No summary reported.")
                cause = results.get("cause", "No cause reported.")
                error = results.get("error", "No error reported.")
                raise ValueError(
                    f"Failed to completed Databricks sync: summary: {summary}, error: {error}, cause: {cause}"
                )
            elif state == "FINISHED":
                return

            raise ValueError("Failed to complete Databricks sync.")

        finally:
            if rendered_template_path and os.path.isfile(rendered_template_path):
                try:
                    os.remove(rendered_template_path)
                except:
                    pass

            if context_id:
                res = self._rest_api_call(
                    "/1.2/contexts/destroy",
                    {"contextId": context_id, "clusterId": cluster_id},
                )

    def _parse_sync_task_state(self, output: dict) -> str:
        metadata = output.get("metadata", {}) or dict()
        tasks = metadata.get("tasks", []) or []
        for task in tasks:
            if task.get("task_key") == SYNC_JOB_TASK_KEY:
                return task.get("state", {}).get("result_state", "PENDING")
        return "PENDING"

    def _create_databricks_cluster(self, config: dict) -> str:
        res = self._rest_api_call("/2.0/clusters/create", config)
        logging.info("start new databricks cluster res: ", res)
        cluster_id = res.get("cluster_id", "")
        if not cluster_id or res.get("error"):
            raise ValueError(res.get("error") or "could not create Databrickc cluster")
        state = ""
        start_time = time.time()
        while state not in (
            "RUNNING",
            "TERMINATING",
            "TERMINATED",
        ):
            res = self._rest_api_call(
                "/2.0/clusters/get", params={"cluster_id": cluster_id}
            )
            state = res.get("state")
            logging.info(f"Waiting on cluster to start. State = {state}")
            time.sleep(5)
        logging.info(f"Finished in {time.time() - start_time} seconds.")
        if state == "RUNNING":
            return cluster_id
        else:
            raise ValueError(res.get("error") or "could not create Databricks cluster")

    def _install_continual_pypi_library(self, cluster_id: str):
        found = False
        res = self._rest_api_call(
            "/2.0/libraries/cluster-status",
            params=dict(cluster_id=cluster_id),
            method="GET",
        )
        logging.info(f"_install_continual_pypi_library uninstall res: {res}")
        library_statuses = res.get("library_statuses", [])
        found = False
        for lib in library_statuses:
            if lib.get("library", {}).get("pypi", {}).get("package", "") == "continual":
                found = True

        if found:
            logging.info("found old version of the continual pypi package")
            res = self._rest_api_call(
                "/2.0/libraries/uninstall",
                {
                    "cluster_id": cluster_id,
                    "libraries": [
                        {"pypi": {"package": "continual==0.5.74"}},
                        {"pypi": {"package": "continual"}},
                    ],
                },
            )
            logging.info(f"uninstalling old continual pypi package res: {res}")

            running = False
            while not running:
                resp = self._rest_api_call(
                    "/2.0/clusters/restart",
                    body={"cluster_id": cluster_id},
                    method="POST",
                )

                resp = self._rest_api_call(
                    "/2.0/clusters/get", body={"cluster_id": cluster_id}
                )
                logging.info("Waiting for cluster to start ...")
                while resp.get("state") != "RUNNING":
                    time.sleep(5)
                    resp = self._rest_api_call(
                        "/2.0/clusters/get", body={"cluster_id": cluster_id}
                    )
                    logging.info(resp.get("state"))

                logging.debug("Successfully revived cluster.")
                running = True

        logging.info("successfully uninstalled existing continual pypi package")

        res = self._rest_api_call(
            "/2.0/libraries/install",
            {
                "cluster_id": cluster_id,
                "libraries": [{"pypi": {"package": "continual"}}],
            },
        )
        logging.info(f"_install_continual_pypi_library res: {res}")

        state = ""
        start = time.time()
        lib = dict()
        while state not in ("INSTALLED", "SKIPPED", "FAILED", "UNINSTALL_ON_RESTART"):
            elapsed = time.time() - start
            if elapsed >= (60 * 10):
                raise ValueError(
                    "Installing Continual PyPI package on Databricks cluster timed out (>15m)."
                )

            res = self._rest_api_call(
                "/2.0/libraries/cluster-status",
                params=dict(cluster_id=cluster_id),
                method="GET",
            )
            logging.info(f"_install_continual_pypi_library install res: {res}")
            library_statuses = res.get("library_statuses", [])
            for lib in library_statuses:
                if (
                    lib.get("library", {}).get("pypi", {}).get("package", "")
                    == "continual"
                ):
                    state = lib.get("status", "").upper()
            time.sleep(5)

        logging.info(
            f"Completed install wait for Continual PyPI package on Databricks cluster: {res}"
        )
        logging.info(
            f"Final state for Continual PyPI package on Databricks cluster: {state}"
        )

        if state == "INSTALLED":
            return

        raise ValueError(
            " ".join(lib.get("messages", []))
            if lib.get("messages")
            else "Failed to install Continual PyPI package on Databricks cluster."
        )

    @handle_databricks_exception
    def _rest_api_call(
        self,
        resource: str,
        body: dict = {},
        method: str = "POST",
        headers: dict = {},
        files: dict = {},
        params: dict = {},
    ) -> dict:
        api_url = "https://{}/api/".format(self.data_store.databricks.host)
        if resource.startswith("/"):
            resource = resource[1:]
        api_url += resource
        headers.update(
            {
                "Authorization": "Bearer {}".format(self.data_store.databricks.token),
                "User-Agent": CONTINUAL_USER_AGENT,
            }
        )
        if method == "GET":
            response = requests.get(api_url, headers=headers, params=params, json=body)
        elif method == "PATCH":
            response = requests.patch(
                api_url, headers=headers, params=params, json=body
            )
        else:
            response = requests.post(
                api_url, headers=headers, params=params, json=body, files=files
            )
        return response.json()

    @handle_databricks_exception
    def _dbfs_rpc(self, action: str, body: dict) -> dict:
        """
        A helper function to make the DBFS API request, request/response is encoded/decoded as JSON.
        Credit: https://docs.databricks.com/dev-tools/api/latest/examples.html#dbfs-large-files
        """
        response = requests.post(
            "https://{}/api/2.0/dbfs/{}".format(
                self.data_store.databricks.host, action
            ),
            headers={
                "Authorization": "Bearer {}".format(self.data_store.databricks.token),
                "User-Agent": CONTINUAL_USER_AGENT,
            },
            json=body,
        )
        return response.json()

    @handle_databricks_exception
    def _delete_file(self, file_path: str) -> bool:
        try:
            self._dbfs_rpc("delete", {"path": file_path})
            return True
        except:
            return False

    @handle_databricks_exception
    def _dbfs_rpc_get(self, action: str, params: dict) -> dict:
        """
        A helper function to make the DBFS API request, request/response is encoded/decoded as JSON.
        Credit: https://docs.databricks.com/dev-tools/api/latest/examples.html#dbfs-large-files
        """
        response = requests.get(
            "https://{}/api/2.0/dbfs/{}".format(
                self.data_store.databricks.host, action
            ),
            headers={
                "Authorization": "Bearer {}".format(self.data_store.databricks.token),
                "User-Agent": CONTINUAL_USER_AGENT,
            },
            params=params,
        )
        return response.json()

    @handle_databricks_exception
    def create_model(self, model):
        response = requests.post(
            "https://{}/api/2.0/preview/mlflow/registered-models/create".format(
                self.data_store.databricks.host
            ),
            headers={
                "Authorization": "Bearer {}".format(self.data_store.databricks.token),
                "User-Agent": CONTINUAL_USER_AGENT,
            },
            json={"name": model.schema.name},
        )

        response = requests.patch(
            "https://{}/api/2.0/preview/mlflow/registered-models/update".format(
                self.data_store.databricks.host
            ),
            headers={
                "Authorization": "Bearer {}".format(self.data_store.databricks.token),
                "User-Agent": CONTINUAL_USER_AGENT,
            },
            json={
                "name": model.schema.name,
                "description": dedent(
                    """
                # ![](https://assets.website-files.com/608c6fda09a6be56ba6844df/608c6fda09a6befa68684532_Symbol.svg)&nbsp;&nbsp;&nbsp;&nbsp;Declaratively Managed by Continual.ai 

                ## Description

                {description}

                ## Documentation

                {documentation}

                ## Links

                Model: {model_link}
                Edit Model: {edit_model_link}

                ## Current YAML

                ```yaml
                {model_yaml}
                ```
                """
                ).format(
                    description=model.schema.description
                    or "_{}_".format("No description provided."),
                    documentation=model.schema.documentation
                    or "_{}_".format("No documentation provided."),
                    model_link="https://cloud.continual.ai/{}".format(model.name),
                    edit_model_link="https://cloud.continual.ai/{}/edit".format(
                        model.name
                    ),
                    model_yaml=model.schema_text,
                ),
            },
        )
        return response.json()

    @handle_databricks_exception
    def create_model_version(self, model: Model, version_id: str):
        try:
            model_version_name = f"{model.name}/versions/{version_id}"
            self._run_databricks_sync(
                "load_model_artifact", {"model_version_name": model_version_name}
            )
            source = (
                f"dbfs:/FileStore/continual/{model_version_name}/model_{version_id}"
            )
            response = requests.post(
                "https://{}/api/2.0/preview/mlflow/model-versions/create".format(
                    self.data_store.databricks.host
                ),
                headers={
                    "Authorization": "Bearer {}".format(
                        self.data_store.databricks.token
                    ),
                    "User-Agent": CONTINUAL_USER_AGENT,
                },
                json={"name": model.schema.name, "source": source},
            )

            model_version = response.json() or dict()

            logging.info(f"created databricks model version: {model_version}")

            update_description_body = {
                "name": model.schema.name,
                "description": dedent(
                    """
                # ![](https://assets.website-files.com/608c6fda09a6be56ba6844df/608c6fda09a6befa68684532_Symbol.svg)&nbsp;&nbsp;&nbsp;&nbsp;Declaratively Managed by Continual.ai 

                ## Description

                {description}

                ## Documentation

                {documentation}

                ## Links

                Model: {model_link}
                Edit Model: {edit_model_link}
                Model Version: {model_version_link}

                ## Current YAML

                ```yaml
                {model_yaml}
                ```
                """
                ).format(
                    description=model.schema.description
                    or "_{}_".format("No description provided."),
                    documentation=model.schema.documentation
                    or "_{}_".format("No documentation provided."),
                    model_link="https://cloud.continual.ai/{}".format(model.name),
                    edit_model_link="https://cloud.continual.ai/{}/edit".format(
                        model.name
                    ),
                    model_version_link="https://cloud.continual.ai/{}/versions/{}".format(
                        model.name, version_id
                    ),
                    model_yaml=model.schema_text,
                ),
            }

            if model_version.get("model_version", dict()).get("version"):
                model_version_payload = update_description_body
                model_version_payload["version"] = model_version["model_version"][
                    "version"
                ]
                response = requests.patch(
                    "https://{}/api/2.0/preview/mlflow/model-versions/update".format(
                        self.data_store.databricks.host
                    ),
                    headers={
                        "Authorization": "Bearer {}".format(
                            self.data_store.databricks.token
                        ),
                        "User-Agent": CONTINUAL_USER_AGENT,
                    },
                    json=model_version_payload,
                )

            logging.info(f"update model version description response: {response}")

            response = requests.patch(
                "https://{}/api/2.0/preview/mlflow/registered-models/update".format(
                    self.data_store.databricks.host
                ),
                headers={
                    "Authorization": "Bearer {}".format(
                        self.data_store.databricks.token
                    ),
                    "User-Agent": CONTINUAL_USER_AGENT,
                },
                json=update_description_body,
            )
            return response.json()
        except Exception:
            raise

    @handle_databricks_exception
    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.info(
            "in write_prediction for [{}]:[{}]".format(
                job_name, job_start_time.strftime("%Y-%m-%d %H:%M:%S")
            )
        )
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )
        df["prediction_time"] = job_start_time.strftime("%Y-%m-%dT%H:%M:%S")

        tab = self.get_prediction_table(model)
        if tab is None:
            print("Prediction table not found")
            logging.debug("Prediction table not found")
            return

        # set chained assignent to warn in case an underlying model has set to raise
        # TODO: determine if there's a better method of mapping the json values
        pd.set_option("mode.chained_assignment", "warn")

        prediction_details_column = f"{output_name}_prediction_details"

        col_names, type_map = self.get_column_names(self.db_schema, "", pred_table)
        existing_col_order = df.columns.tolist()
        sorted_col_names = list(filter(lambda c: c in existing_col_order, col_names))
        df = df[sorted_col_names]

        if prediction_details_column in df.columns:
            df[prediction_details_column] = df[prediction_details_column].map(
                json.dumps
            )

        self._two_stage_insert(df, pred_table, exists="append")

        count_query = f"SELECT COUNT(1) from {self.db_schema}.{pred_table} WHERE batch_prediction = '{job_name}'"
        result = self.connection.execute(count_query)
        for row in result:
            return row[0]
        return 0

    @handle_databricks_exception
    def _two_stage_insert(
        self,
        df: pd.DataFrame,
        table_name: str,
        exists: str = "fail",
    ):
        ch = chunk_it(0, df.shape[0], step=10_000)
        while ch is not None:
            df_slice = df[ch[0] : ch[1]]
            logging.info("writing features to table {}".format(table_name))
            logging.info("feature dtypes: {}".format(df_slice.dtypes))

            try:
                chunk_start = time.time()
                logging.info(
                    "Writing chunk ({} to {}) of DataFrame ...".format(ch[0], ch[1])
                )
                df_slice.to_sql(
                    table_name,
                    self.connection,
                    schema=self.db_schema,
                    if_exists=exists,
                    method="multi",
                    index=False,
                )
                logging.info(
                    "Done. Chunk took {}s to write.".format(time.time() - chunk_start)
                )
                if ch[0] == 0:
                    logging.info("Converting to Delta table ...")
                    delta_start = time.time()
                    self.connection.execute(
                        "CONVERT TO DELTA {}.{} NO STATISTICS".format(
                            self.db_schema, table_name
                        )
                    )
                    logging.info(
                        "Done converting into Delta. Took {}s".format(
                            time.time() - delta_start
                        )
                    )
            except:
                traceback.print_exc()
                logging.debug("write failed to execute - aborting")
                raise
            ch = chunk_it(ch[1], df.shape[0], step=10_000)

    @handle_databricks_exception
    def _create_prediction_model_table(
        self, model, model_pred_name, meta: MetaData, problem_type
    ):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """

        index = None
        time_index = None
        target_column = model.schema.target
        target_column_dtype = db.TEXT
        for f in model.schema.columns:
            if (
                f.type == management_types_pb2.FieldType.INDEX
                or f.name == model.schema.index
            ):
                index = f
            elif (
                f.type == management_types_pb2.FieldType.TIME_INDEX
                or f.name == model.schema.time_index
            ):
                time_index = f
            elif f.name == target_column:
                target_column_dtype = Typemap[f.dtype]

        columns = []
        columns.append(db.Column("batch_prediction", db.Text))
        columns.append(db.Column("model_version", db.Text))
        columns.append(db.Column("prediction_time", db.TIMESTAMP()))
        columns.append(db.Column(f"features", db.Text))

        if index is not None:
            columns.append(db.Column(index.name, Typemap[index.dtype]))

        if time_index is not None:
            columns.append(db.Column(time_index.name, Typemap[time_index.dtype]))
        columns.append(db.Column(f"{target_column}_prediction", target_column_dtype))
        if problem_type == "multiclass_classification":
            columns.append(db.Column(f"{target_column}_prediction_details", db.Text))
            columns.append(db.Column(f"{target_column}_prediction_score", db.Float))
        if problem_type == "binary_classification":
            columns.append(
                db.Column(f"{target_column}_true_prediction_score", db.Float)
            )

        table = db.Table(model_pred_name, meta, *columns)
        logging.info("databricks model prediction table: {}".format(table))

        # Following retry logic is added in case a parallel operation
        # creates a schema that might not have existed at the start of the session.
        count = 0
        done = False
        while not done:
            try:
                table.create(self.engine, checkfirst=False)
                # meta.create_all(self.engine, checkfirst=False)
                done = True
            except (IntegrityError, ProgrammingError):
                logging.debug(traceback.format_exc())
                logging.debug("Error while creating model table " + str(count))
                count = count + 1
                if count == 2:
                    raise

        self.connection.execute(
            "CONVERT TO DELTA {}.{} NO STATISTICS".format(
                self.db_schema, model_pred_name
            )
        )

    @handle_databricks_exception
    def load_data(self, filename, table_name, replace=False):
        source = FeatureSourceCloudStore(str(uuid.uuid4()), filename)
        df = source.fetch_all()
        df.columns = [col.lower() for col in df.columns]
        exists = "fail"
        if replace:
            exists = "replace"
        try:
            logging.info(
                "load_data: schema: {}, table_name: {}".format(
                    self.db_schema, table_name.lower()
                )
            )
            self._two_stage_insert(df, table_name.lower(), exists=exists)
        except ValueError:
            traceback.print_exc()

        return table_name.lower()

    @handle_databricks_exception
    def infer_schema(
        self,
        query,
        excludes=None,
        sample_row_count=100,
        allow_invalid_types=False,
        timeout_ms=0,
    ):
        return super().infer_schema(
            query,
            excludes=excludes,
            sample_row_count=sample_row_count,
            allow_invalid_types=allow_invalid_types,
            timeout_ms=timeout_ms,
        )



import grpc
import collections
from typing import Callable


class ClientCallDetails(
    collections.namedtuple(
        "ClientCallDetails", ("method", "timeout", "metadata", "credentials")
    ),
    grpc.ClientCallDetails,
):
    """Concrete client call details object.

    grpc.ClientCallDetails is an abstract base class.
    """


class AuthInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Adds API key to GRPC requests."""

    api_key_getter: Callable[[], str]

    def __init__(self, api_key_getter: Callable[[], str]):
        """Initialize authentication interceptor.

        This class takes a getter callable in order to support
        lazy loading of the api_key.

        Arguments:
            api_key_getter: Function that returns an api key.
        """
        self.api_key_getter = api_key_getter

    def intercept_unary_unary(self, continuation, client_call_details, request):
        api_key = self.api_key_getter()
        if api_key is not None and api_key != "":
            metadata = [("authorization", "Bearer " + api_key)]
        else:
            metadata = []
        client_call_details = ClientCallDetails(
            method=client_call_details.method,
            metadata=metadata,
            timeout=client_call_details.timeout,
            credentials=client_call_details.credentials,
        )
        return continuation(client_call_details, request)



import logging

import sqlalchemy as db
from sqlalchemy.exc import NoSuchTableError

from datetime import datetime
import json
import os
import traceback
import pandas as pd

from .query_templates import stats_template
from .featurestore import FeatureStore
from .prediction_store import PredictionStore, chunk_it


class PostgresFeatureStore(FeatureStore, PredictionStore):
    def __init__(self, data_store, source, cfg):
        self.data_store = data_store
        db_url = "postgresql://continual:continual@localhost:5432/continual"
        logging.info(f"PostgresFeatureStore data_store: {data_store}")
        try:
            if (
                data_store is None
                or data_store.postgres is None
                or data_store.postgres.host == ""
            ):
                if (
                    os.environ.get("CONFIG_ENV", "dev") != "test"
                    and "DatabaseURL" in cfg
                    and cfg["DatabaseURL"] is not None
                ):
                    db_url = cfg["DatabaseURL"]
            else:
                config = data_store.postgres
                db_url = "postgresql://{}:{}@{}:5432/{}".format(
                    config.username,
                    config.password,
                    config.host,
                    config.database,
                )
        except Exception as e:
            logging.error(
                "encountered error in postgres featurestore init: {}".format(e)
            )

        self.engine = db.create_engine(db_url)
        self.connection = self.engine.connect()
        self.trans = None
        super(PostgresFeatureStore, self).__init__()

    def create_feature_table(self, feature_set):
        """
         Creates featuretable.
         This does not check for the existance of the featuretable and should
           be done prior to calling this.

        Args: Feature set for which the tables are to be created.≈
        Returns None
        """

        return super().create_feature_table(feature_set)

    def get_feature_table(self, feature_set):
        project, tablename = self.get_table_name(feature_set)
        tablename = "featureset_{}".format(tablename)

        try:
            tab = db.Table(
                tablename,
                db.MetaData(schema=self.db_schema),
                autoload=True,
                autoload_with=self.engine,
            )
            return tab
        except NoSuchTableError:
            print("No table found error")
            return None

    def set_db_schema(self, db_schema):
        self.db_schema = db_schema
        self.connection.execute("SET search_path TO {},public".format(db_schema))

    def cleanup(self):
        self.connection.execute("DROP SCHEMA IF EXISTS " + self.db_schema + " CASCADE")
        return

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None

    def get_type(self) -> str:
        return "postgres"

    def is_hosted(self):
        return True

    def get_stats_query_template(self):
        return stats_template

    def get_database_names(self):
        raise NotImplementedError()

    def get_schema_names(self, database):
        raise NotImplementedError()

    def get_table_names(self, database, schema):
        raise NotImplementedError()

    def get_column_names(self, database, schema, table):
        raise NotImplementedError()

    def get_prediction_table(self, model):
        project_id, pred_table = self.get_model_prediction_table_name(model)
        print("Prediction table in " + self.db_schema)
        try:
            tab = db.Table(
                pred_table,
                db.MetaData(schema=self.db_schema),
                autoload=True,
                autoload_with=self.engine,
            )
            return tab
        except NoSuchTableError:
            # Todo perhaps not capture the error here.
            return None

    def write_prediction(
        self,
        df: pd.DataFrame,
        id_column_name: str,
        ts_column_name: str,
        model_query: any,
        output_name: str,
        job_name: str,
        job_start_time: datetime,
        model,
        model_version,
        time_index_dtype,
    ):
        logging.debug(
            "in write_prediction for [{}]:[{}]".format(
                job_name, job_start_time.strftime("%Y-%m-%d %H:%M:%S")
            )
        )
        if self.data_store is None:
            logging.debug("skipping write_prediction")
            return

        project_id, pred_table = self.get_model_prediction_table_name(model)
        df = self._process_dataframe(
            df, id_column_name, ts_column_name, job_name, model_version
        )
        df["prediction_time"] = job_start_time.strftime("%Y-%m-%d %H:%M:%S")

        tab = self.get_prediction_table(model)
        if tab is None:
            print("Prediction table not found")
            logging.debug("Prediction table not found")
            return

        # set chained assignent to warn in case an underlying model has set to raise
        # TODO: determine if there's a better method of mapping the json values
        pd.set_option("mode.chained_assignment", "warn")
        trans = self.connection.begin()

        prediction_details_column = f"{output_name}_prediction_details"

        ch = chunk_it(0, df.shape[0])
        while ch is not None:
            df_slice = df[ch[0] : ch[1]]
            if prediction_details_column in df_slice.columns:
                df_slice[prediction_details_column] = df_slice[
                    prediction_details_column
                ].map(json.dumps)
            df_slice["features"] = df_slice["features"].map(json.dumps)

            try:
                df_slice.to_sql(
                    pred_table,
                    self.connection,
                    schema=self.db_schema,
                    if_exists="append",
                    index=False,
                )
            except:
                trans.rollback()
                traceback.print_exc()
                logging.debug("write failed to execute - aborting")
                raise
            ch = chunk_it(ch[1], df.shape[0])

        trans.commit()

        count_query = f"SELECT COUNT(1) from {self.db_schema}.{pred_table} WHERE batch_prediction = '{job_name}'"
        result = self.connection.execute(count_query)
        for row in result:
            return row[0]
        return 0



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/rpc/error_details.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/rpc/error_details.proto',
  package='continual.rpc.rpc',
  syntax='proto3',
  serialized_options=b'Z6github.com/continualiq/continual/continual/rpc/rpc;rpc',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%continual/rpc/rpc/error_details.proto\x12\x11\x63ontinual.rpc.rpc\"}\n\x0c\x45rrorDetails\x12=\n\x07\x64\x65tails\x18\x01 \x03(\x0b\x32,.continual.rpc.rpc.ErrorDetails.DetailsEntry\x1a.\n\x0c\x44\x65tailsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x38Z6github.com/continualiq/continual/continual/rpc/rpc;rpcb\x06proto3'
)




_ERRORDETAILS_DETAILSENTRY = _descriptor.Descriptor(
  name='DetailsEntry',
  full_name='continual.rpc.rpc.ErrorDetails.DetailsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.rpc.ErrorDetails.DetailsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.rpc.ErrorDetails.DetailsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=185,
)

_ERRORDETAILS = _descriptor.Descriptor(
  name='ErrorDetails',
  full_name='continual.rpc.rpc.ErrorDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='details', full_name='continual.rpc.rpc.ErrorDetails.details', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ERRORDETAILS_DETAILSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=185,
)

_ERRORDETAILS_DETAILSENTRY.containing_type = _ERRORDETAILS
_ERRORDETAILS.fields_by_name['details'].message_type = _ERRORDETAILS_DETAILSENTRY
DESCRIPTOR.message_types_by_name['ErrorDetails'] = _ERRORDETAILS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ErrorDetails = _reflection.GeneratedProtocolMessageType('ErrorDetails', (_message.Message,), {

  'DetailsEntry' : _reflection.GeneratedProtocolMessageType('DetailsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ERRORDETAILS_DETAILSENTRY,
    '__module__' : 'continual.rpc.rpc.error_details_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.rpc.ErrorDetails.DetailsEntry)
    })
  ,
  'DESCRIPTOR' : _ERRORDETAILS,
  '__module__' : 'continual.rpc.rpc.error_details_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.rpc.ErrorDetails)
  })
_sym_db.RegisterMessage(ErrorDetails)
_sym_db.RegisterMessage(ErrorDetails.DetailsEntry)


DESCRIPTOR._options = None
_ERRORDETAILS_DETAILSENTRY._options = None
# @@protoc_insertion_point(module_scope)
